import importlib.metadata

from ghq.univariate import univariate
from ghq.univariate import univariate_importance

from ghq.multivariate import multivariate
from ghq.multivariate import multivariate_importance
from ghq.multivariate import _multivariate_cov_sqrt
from ghq.multivariate import multivariate_normal_log_pdf

__version__ = importlib.metadata.version("ghq")

del importlib



from typing import Callable
from itertools import product
from functools import partial

from numpy.polynomial.hermite_e import hermegauss
from jax import numpy as jnp, jit, Array, vmap
from jax.scipy.linalg import cho_factor, solve_triangular


@partial(jit, static_argnums=(0, 3))
def multivariate(
    integrand: Callable[[Array], Array],
    mean: Array,
    cov: Array,
    degree: int = 32,
) -> float:
    """
    Multivariate Gauss-Hermite integration.
    https://en.wikipedia.org/wiki/Gauss-Hermite_quadrature

    Not vectorised in mean and cov, but works with jax.vmap.

    .. math::

        \int integrand(x) \text{N}(x | mean, cov) dx

    Args:
        integrand: Function to be integrated over f(x).
        mean: Mean of multivariate Gaussian.
        cov: Covariance matrix of multivariate Gaussian.
        degree: Integer number of Gauss-Hermite points, defaults to 32.
    Returns:
        Approximate Gaussian expectation.
    """
    cov_sqrt, _ = cho_factor(cov, lower=True)
    return _multivariate_cov_sqrt(integrand, mean, cov_sqrt, degree)


def _multivariate_cov_sqrt(
    integrand: Callable[[Array], Array],
    mean: Array,
    cov_sqrt: Array,
    degree: int,
) -> float:
    dim = mean.size

    def standardized_integrand(x):
        return integrand(mean + cov_sqrt @ x)

    x, w = hermegauss(degree)

    xn = jnp.array(list(product(*(x,) * dim)))
    wn = jnp.prod(jnp.array(list(product(*(w,) * dim))), 1)

    hx = vmap(standardized_integrand)(xn)
    wn = wn.reshape((degree**dim,) + (1,) * (hx.ndim - 1))
    return (wn * hx).sum(0) / jnp.sqrt(2 * jnp.pi) ** dim


@partial(jit, static_argnums=(0, 3))
def multivariate_importance(
    integrand: Callable[[Array], Array],
    mean: Array,
    cov: Array = None,
    degree: int = 32,
    lower: list | tuple | Array = None,
    upper: list | tuple | Array = None,
) -> Array:
    """
    Numerical integration using Gauss-Hermite over a multivariate Gaussian(s)
    with importance corrected weights.

    Not vectorised in mean and cov, but works with jax.vmap.

    .. math::

        \int integrand(x)dx
            = \int integrand(x) / N(x | mean, cov) * N(x | mean, cov) dx

    Args:
        integrand: Function to be integrated over f(x), vectorised in x.
        mean: Mean of multivariate Gaussian (dim,).
        cov: Covariance matrix of multivariate Gaussian (dim, dim).
        degree: Integer number of Gauss-Hermite points, defaults to 32.
        lower: Lower bounds of integration, defaults to -inf.
        upper: Upper bounds of integration, defaults to inf.
    Returns:
        Approximate integration value.
    """

    if lower is not None or upper is not None:
        raise NotImplementedError(
            "Constrained multivariate integration not implemented yet."
        )

    dim = mean.size

    if cov is None:
        cov = jnp.eye(dim)

    cov_sqrt, _ = cho_factor(cov, lower=True)

    def importance_integrand(x):
        pdf_evals = jnp.exp(multivariate_normal_log_pdf(x, mean, cov_sqrt))
        return jnp.where(pdf_evals > 0, integrand(x) / pdf_evals, 0)

    return _multivariate_cov_sqrt(importance_integrand, mean, cov_sqrt, degree)


def multivariate_normal_log_pdf(x, mean, cov_sqrt):
    """
    Log PDF of a multivariate Gaussian.

    Vectorized in x, not in mean and cov_sqrt. But works with jax.vmap.

    Args:
        x: Array of multivariate Gaussian samples (n, dim).
        mean: Mean of multivariate Gaussian (dim,).
        cov_sqrt: Lower triangular Cholesky decomposition
            of covariance matrix (dim, dim).

    Returns:
        Array of log PDF evaluations (n,).
    """

    dim = mean.size

    centered_x = x - mean
    y = solve_triangular(cov_sqrt, centered_x.T, lower=True).T
    maha = jnp.sum(y**2, axis=-1)
    log_det = 2 * jnp.sum(jnp.log(jnp.diag(cov_sqrt)))
    norm_const = dim * jnp.log(2 * jnp.pi) + log_det
    log_pdf = -0.5 * (norm_const + maha)
    return log_pdf



from jax import numpy as jnp, jit, vmap
from jax.scipy.stats import multivariate_normal

import ghq


def test_zero():
    def f(x):
        return 0.0

    mean = jnp.array([0.0])
    cov = jnp.array([[1.0]])
    degree = 32
    result = jit(ghq.multivariate, static_argnums=(0, 3))(f, mean, cov, degree)
    assert jnp.isclose(result, 0.0)
    importance_result = jit(ghq.multivariate_importance, static_argnums=(0, 3))(
        f, mean, cov, degree
    )
    assert jnp.isclose(importance_result, 0.0)

    mean = jnp.array([0.0, 0.0])
    cov = jnp.array([[1.0, -0.3], [-0.3, 1.0]])
    degree = 32
    result = ghq.multivariate(f, mean, cov, degree)
    assert jnp.isclose(result, 0.0)
    importance_result = jit(ghq.multivariate_importance, static_argnums=(0, 3))(
        f, mean, cov, degree
    )
    assert jnp.isclose(importance_result, 0.0)


def test_x0():
    def f(x):
        return x[0]

    mean = jnp.array([0.0])
    cov = jnp.array([[1.0]])
    degree = 32
    result = ghq.multivariate(f, mean, cov, degree)
    assert jnp.isclose(result, 0.0)

    mean = jnp.array([1.0, 0.0])
    cov = jnp.array([[1.0, -0.3], [-0.3, 1.0]])
    degree = 32
    result = ghq.multivariate(f, mean, cov, degree)
    assert jnp.isclose(result, 1.0)


def test_corr():
    def f(x):
        return x[0] * x[1]

    mean = jnp.array([0.0, 0.0])
    cov = jnp.array([[1.0, -0.3], [-0.3, 1.0]])
    degree = 32
    result = ghq.multivariate(f, mean, cov, degree)
    assert jnp.isclose(result, -0.3)


def test_matrix_integrand():
    def f(x):
        return jnp.outer(x, x)

    mean = jnp.array([0.0, 0.0, 0.0])
    cov = jnp.array([[1.0, -0.3, 0.2], [-0.3, 1.0, 0.1], [0.2, 0.1, 1.0]])
    degree = 32
    result = ghq.multivariate(f, mean, cov, degree)
    assert jnp.allclose(result, cov)


def test_multivariate_normal_log_pdf():
    mean = jnp.array([-0.2, 3.0])
    cov = jnp.array([[1.0, -0.3], [-0.3, 1.0]])
    x = jnp.arange(8).reshape(4, 2)
    cov_sqrt = jnp.linalg.cholesky(cov)
    result = ghq.multivariate_normal_log_pdf(x, mean, cov_sqrt)
    expected = vmap(multivariate_normal.logpdf, in_axes=(0, None, None))(x, mean, cov)
    assert jnp.allclose(result, expected)


def test_importance_exp_decay():
    def f(x):
        return jnp.exp(-jnp.abs(x).sum())

    mean = jnp.array([1.0, 0.3])
    cov = jnp.array([[1.0, -0.3], [-0.3, 1.0]])
    degree = 64
    result = ghq.multivariate_importance(f, mean, cov, degree)
    assert jnp.isclose(result, 4.0, atol=1e-1)


def test_importance_gaussian():
    def f(x):
        return jnp.exp(-jnp.sum(x**2) / 2)

    mean = jnp.array([-0.2, 0.8, 0.3])
    cov = jnp.array([[0.8, 0.5, 0.1], [0.5, 1.1, -0.3], [0.1, -0.3, 1.0]])
    degree = 64
    result = ghq.multivariate_importance(f, mean, cov, degree)
    assert jnp.isclose(result, (2 * jnp.pi) ** 1.5, atol=1e-1)



from typing import Callable
from functools import partial

from numpy.polynomial.hermite_e import hermegauss
from jax import numpy as jnp, jit, vmap, Array
from jax.lax import cond
from jax.scipy.stats import norm


def logit(u: float) -> float:
    return jnp.log(u / (1 - u))


def inverse_logit(x: float) -> float:
    return 1 / (1 + jnp.exp(-x))


@partial(jit, static_argnums=(0, 3))
def univariate(
    integrand: Callable[[Array], Array],
    mean: Array,
    sd: Array,
    degree: int = 32,
) -> Array:
    """
    Gauss-Hermite integration over 1-D Gaussian(s).
    https://en.wikipedia.org/wiki/Gauss-Hermite_quadrature

    .. math::

        \int integrand(x) \text{N}(x | mean, sd^2) dx

    Args:
        integrand: Function to be integrated over f(x), vectorised.
        mean: Array of n means each corresponding to a 1-D Gaussian (n,).
        sd: Array of n standard deviations each corresponding to a 1-D Gaussian (n,).
        degree: Integer number of Gauss-Hermite points, defaults to 32.
    Returns:
        Array of n approximate 1D Gaussian expectations.
    """
    mean = jnp.array(mean)
    sd = jnp.array(sd)

    n = mean.size
    x, w = hermegauss(degree)
    x = jnp.repeat(x[..., jnp.newaxis], n, axis=1)  # extend shape to (degree, n)
    x = sd * x + mean
    hx = vmap(integrand)(x)
    w = w.reshape((degree,) + (1,) * (hx.ndim - 1))
    return jnp.squeeze((w * hx).sum(0) / jnp.sqrt(2 * jnp.pi))


def lower_bound_transform(y, integrand, lower):
    ey = jnp.exp(y)
    return integrand(lower + ey) * ey


def upper_bound_transform(y, integrand, upper):
    ey = jnp.exp(y)
    return -integrand(upper - ey) * ey


def bounded_transform(y, integrand, lower, upper):
    ily = inverse_logit(y)
    return integrand(lower + (upper - lower) * ily) * (upper - lower) * ily * (1 - ily)


@partial(jit, static_argnums=(0, 3))
def univariate_importance(
    integrand: Callable[[Array], Array],
    mean: Array = jnp.zeros(1),
    sd: Array = jnp.ones(1),
    degree: int = 32,
    lower: float = -jnp.inf,
    upper: float = jnp.inf,
) -> Array:
    """
    Numerical integration using Gauss-Hermite over a 1-D Gaussian(s)
    with importance corrected weights.

    .. math::

        \int integrand(x)dx
            = \int integrand(x) / N(x | mean, sd^2) * N(x | mean, sd^2) dx

    Args:
        integrand: Function to be integrated over f(x), vectorised in x.
        mean: Array of n means each corresponding to a 1-D Gaussian (n,), defaults to 0.
        sd: Array of n standard deviations each corresponding to a 1-D Gaussian (n,),
            defaults to 1.
        degree: Integer number of Gauss-Hermite points, defaults to 32.
        lower: Lower bound of integration, defaults to -inf.
        upper: Upper bound of integration, defaults to inf.
    Returns:
        Array of n approximate 1D integrations.
    """

    def transformed_integrand(y):
        res = cond(
            (lower != -jnp.inf) * (upper == jnp.inf),
            lambda z: lower_bound_transform(z, integrand, lower),
            lambda _: y,
            y,
        )

        res = cond(
            (lower == -jnp.inf) * (upper != jnp.inf),
            lambda z: upper_bound_transform(z, integrand, upper),
            lambda _: y,
            y,
        )

        res = cond(
            (lower != -jnp.inf) * (upper != jnp.inf),
            lambda z: bounded_transform(z, integrand, lower, upper),
            lambda _: y,
            y,
        )
        return res

    def importance_integrand(x):
        pdf_evals = norm.pdf(x, loc=mean, scale=sd)
        return jnp.where(pdf_evals > 0, transformed_integrand(x) / pdf_evals, 0)

    return univariate(importance_integrand, mean, sd, degree)



from jax import numpy as jnp, jit, random

import ghq


def test_zero():
    def f(x):
        return 0.0

    assert jnp.isclose(jit(ghq.univariate, static_argnums=(0,))(f, 0, 1), 0.0)
    assert jnp.isclose(
        jit(ghq.univariate_importance, static_argnums=(0,))(f, 0, 1), 0.0, atol=1e-3
    )
    assert jnp.isclose(
        jit(ghq.univariate_importance, static_argnums=(0,))(f, 0, 1, lower=0),
        0.0,
        atol=1e-3,
    )
    assert jnp.isclose(
        jit(ghq.univariate_importance, static_argnums=(0,))(f, 0, 1, upper=0),
        0.0,
        atol=1e-3,
    )
    assert jnp.isclose(
        jit(ghq.univariate_importance, static_argnums=(0,))(f, 0, 1, lower=0, upper=1),
        0.0,
        atol=1e-3,
    )


def test_x():
    def f(x):
        return x

    assert jnp.isclose(ghq.univariate(f, 0, 1), 0.0)
    assert jnp.isclose(
        ghq.univariate_importance(f, 0, 1, lower=-1, upper=1), 0.0, atol=1e-3
    )


def test_abs_x():
    def f(x):
        return jnp.abs(x)

    assert jnp.isclose(
        ghq.univariate(f, 0, 1, degree=50),
        f(random.normal(key=random.PRNGKey(0), shape=(10000,))).mean(),
        atol=1e-2,
    )
    assert jnp.isclose(
        ghq.univariate_importance(f, 0, 1, lower=-1, upper=1), 1.0, atol=1e-2
    )


def test_x2():
    def f(x):
        return jnp.square(x)

    assert jnp.isclose(ghq.univariate(f, 0, 1), 1.0)
    assert jnp.isclose(
        ghq.univariate_importance(f, 0, 1, lower=-1, upper=1), 2 / 3, atol=1e-2
    )


def test_polynomial():
    def f(x):
        return 0.1 * x**5 - 0.3 * x**3 + 2 * x**2 + 3 * x + 4

    mn = 0.3
    sd = 0.5

    assert jnp.isclose(
        ghq.univariate(f, mn, sd, degree=50),
        f(mn + sd * random.normal(key=random.PRNGKey(0), shape=(10000,))).mean(),
        atol=1e-1,
    )
    assert jnp.isclose(
        ghq.univariate_importance(f, mn, sd, degree=50, lower=-1, upper=1),
        9.33333,
        atol=1e-1,
    )


def test_matrix_integrand():
    def f(x):
        return x**2 * jnp.eye(3)

    assert jnp.allclose(jit(ghq.univariate, static_argnums=(0,))(f, 0, 1), jnp.eye(3))
